# -*- coding: utf-8 -*-
"""ahp.errors

This module contains the errors which are raised by this package.
"""


class AHPModelError(Exception):
    """Base class for exceptions in this module."""
    pass


class AHPMethodUnsupportedError(AHPModelError):
    """Error for cases when the AHP model uses a method not implemented."""
    def __init__(self, expected, actual):
        self.expected = expected
        self.actual = actual

    def __str__(self):
        return 'Expected method to be one of [{}], got \'{}\''.format(self.expected, self.actual)


class AHPFieldEmptyError(AHPModelError):
    """Error for cases when the AHP model has an empty field."""
    def __init__(self, var):
        self.var = var

    def __str__(self):
        return 'Field \'{}\' should not be empty'.format(self.var)


class AHPTypeMismatchError(AHPModelError):
    """Error for cases when the properties in the AHP model don't match the expected type."""
    def __init__(self, var, expected, actual, list_elements=False):
        self.var = var
        self.expected = expected
        self.actual = actual
        self.list_elements = list_elements

    def __str__(self):
        if self.list_elements:
            return 'Expected \'{}\' list to have <{}> elements, got <{}>'.format(self.var, self.expected, self.actual)
        else:
            return 'Expected \'{}\' to be of type <{}>, got <{}>'.format(self.var, self.expected, self.actual)


class AHPContainsDuplicateError(AHPModelError):
    """Error for cases when the properties in the AHP model have duplicate entries."""
    def __init__(self, var):
        self.var = var

    def __str__(self):
        return 'Field \'{}\' contains duplicates'.format(self.var)


class AHPNonSquarePreferenceMatrixError(AHPModelError):
    """Error for cases when the preference matrix in the AHP model are not square matrices."""
    def __init__(self, kind, name, side, actual_width, actual_height):
        self.kind = kind
        self.name = name
        self.side = side
        self.actual_width = actual_width
        self.actual_height = actual_height

    def __str__(self):
        return 'Expecting \'{0}:{1}\' preference matrix to be {2}x{2} got {3}x{4}'.format(self.kind, self.name,
                                                                                          self.side,
                                                                                          self.actual_width,
                                                                                          self.actual_height)


class AHPMissingPreferenceMatrixError(AHPModelError):
    """Error for cases when a required preference matrix is missing in the AHP model."""
    def __init__(self, kind, name):
        self.kind = kind
        self.name = name

    def __str__(self):
        return 'Missing \'{}\' preference matrix for \'{}\''.format(self.kind, self.name)



# -*- coding: utf-8 -*-
"""ahp.hierarchy.ahpmodel

This module contains the class definition for the AHP Model Node (root) in the hierarchy model.
"""

import numpy as np

from firehole.algorithms.ahp.hierarchy import AHPCriterion
from firehole.algorithms.ahp.methods import EigenvalueMethod
from firehole.algorithms.ahp.utils import normalize_priorities


class AHPModel:
    """AHPModel

    Args:
        model (dict): The Analytic Hierarchy Process model.
        solver (ahp.methods): Method used when calculating the priorities of the lower layer.
    """

    def __init__(self, model, solver=EigenvalueMethod):
        self.solver = solver()
        self.preference_matrices = model['preferenceMatrices']

        criteria = model.get('criteria')
        self.criteria = [AHPCriterion(n, model, solver=solver) for n in criteria]

    def get_priorities(self, round_results=True, decimals=3):
        """Get the priority of the nodes in the level below this node.

        Args:
            round_results (bool): Return rounded priorities. Default is True.
            decimals (int): Number of decimals to round to, ignored if `round_results=False`. Default is 3.

        Returns:
            Global priorities of the alternatives in the model, rounded to `decimals` positions if `round_results=True`.
        """
        crit_pm = np.array(self.preference_matrices['criteria'])
        crit_pr = self.solver.estimate(crit_pm)

        crit_attr_pr = [criterion.get_priorities() for criterion in self.criteria]
        priorities = normalize_priorities(crit_attr_pr, crit_pr)

        if round_results:
            return np.around(priorities, decimals=decimals)

        return priorities



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from firehole.algorithms.similarity.rank_bm25 import BM25Okapi, BM25L, BM25Plus
from firehole.algorithms.similarity.simhash import Simhash, SimhashIndex



# -*- coding: utf-8 -*-
"""ahp.methods.eigenvalue

This module contains the class implementing the eigenvalue priority estimation method.
"""

import numpy as np
from scipy.sparse.linalg import eigs

from firehole.algorithms.ahp.methods import Method

RANDOM_INDICES = [0, 0, 0.58, 0.9, 1.12, 1.24, 1.32, 1.41, 1.45, 1.49, 1.51]


class EigenvalueMethod(Method):
    """Eigenvalue based priority estimation method
    """

    @staticmethod
    def _evaluate_consistency(matrix):
        width = matrix.shape[0]

        if width > len(RANDOM_INDICES):
            return 0

        return RANDOM_INDICES[width - 1]

    def estimate(self, preference_matrix):
        super()._check_matrix(preference_matrix)
        width = preference_matrix.shape[0]

        _, vectors = eigs(preference_matrix, k=(width-2), sigma=width, which='LM', v0=np.ones(width))

        real_vector = np.real([vec for vec in np.transpose(vectors) if not np.all(np.imag(vec))][:1])
        sum_vector = np.sum(real_vector)

        self._evaluate_consistency(preference_matrix)

        return np.around(real_vector, decimals=3)[0] / sum_vector



# -*- coding: utf-8 -*-
"""ahp.__main__

This module contains the common functions and methods to provide a CLI for the AHP package.
"""

import argparse
import json

from firehole.algorithms.ahp import parse
from firehole.algorithms.ahp.errors import AHPModelError


def parse_args():
    """Parse command line arguments.

    Returns:
         Command line arguments in form of a Namespace
    """
    parser = argparse.ArgumentParser(description='Solver for Analytic Hierarchy Process models.')
    parser.add_argument('-f', '--file',
                        metavar='FILE',
                        type=str,
                        nargs='+',
                        required=True,
                        help='configuration file(s) for Analytic Hierarchy Process')

    return parser.parse_args()


def print_priorities(alternatives, priorities):
    """Pretty print the alternatives and the priorities.

    Args:
        alternatives (list): List of alternatives.
        priorities (list): List of priorities corresponding to the alternatives.
    """
    print('\tResults:')
    for idx, alternative in enumerate(alternatives):
        print('\t\t{}: {}'.format(alternative, priorities[idx]))


def main():
    """Main function of the script.

    This function parses the command line arguments, validates and creates the AHP models
    and prints the summary of all the AHP models.
    """
    args = parse_args()

    models = {file: json.load(open(file)) for file in args.file}

    for name, model in models.items():
        print('[+] {}'.format(model.get('name', name)))
        try:
            ahp_model = parse(model)
            print('\tMethod: {}'.format(model['method']))

            print_priorities(model['alternatives'], ahp_model.get_priorities())

        except Exception as err:
            print('\t[-] ERROR:{} {}'.format(err.__class__.__name__, err))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""ahp.methods.method

This module contains the abstract Method class used as a base class for other methods.
"""

from abc import ABC, abstractmethod


class Method(ABC):
    """Method

    Abstract class to provide a common interface for various methods.
    """

    @abstractmethod
    def estimate(self, preference_matrix):
        """Estimate the priority from the provided preference matrix

        Args:
            preference_matrix (np.array): A two (nxn) dimensional reciprocal array.

        Returns:
            nx1 np.array of priorities
        """
        pass

    @staticmethod
    def _check_matrix(matrix):
        width, height = matrix.shape

        assert width == height, "Preference Matrix should be a square matrix"
        assert width >= 2, "Preference Matrix too small or empty"

        for i in range(width):
            for j in range(height):
                if i == j:
                    assert matrix[i, j] == 1, "Preference should be 1 on the diagonal"
                else:
                    assert abs(1 - matrix[i, j]*matrix[j, i]) <= 0.011, "Failed consistency check for Reciprocal Matrix"



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# There are import oderwise
from firehole.algorithms.flashtext import *
from firehole.algorithms.weight import *
from firehole.algorithms.ahp import *
from firehole.algorithms.similarity import *

import firehole.algorithms.flashtext
import firehole.algorithms.weight
import firehole.algorithms.ahp
import firehole.algorithms.similarity

# Need to test with Numpy, when available
# weight
from firehole.algorithms.weight import (Entropy, COV)
from firehole.algorithms.ahp import parse

# Keyword extraction and replace
from firehole.algorithms.flashtext import KeywordProcessor

# calculate the text similarity
from firehole.algorithms.similarity import (BM25Plus, BM25L, BM25Okapi)
from firehole.algorithms.similarity import (Simhash, SimhashIndex)



# Created by 1e0n in 2013
from __future__ import division, unicode_literals

import re
import sys
import hashlib
import logging
import numbers
import collections
from itertools import groupby

if sys.version_info[0] >= 3:
    basestring = str
    unicode = str
    long = int
else:
    range = xrange


def _hashfunc(x):
    return int(hashlib.md5(x).hexdigest(), 16)


class Simhash(object):

    def __init__(
            self, value, f=64, reg=r'[\w\u4e00-\u9fcc]+', hashfunc=None, log=None
    ):
        """
        `f` is the dimensions of fingerprints

        `reg` is meaningful only when `value` is basestring and describes
        what is considered to be a letter inside parsed string. Regexp
        object can also be specified (some attempt to handle any letters
        is to specify reg=re.compile(r'\w', re.UNICODE))

        `hashfunc` accepts a utf-8 encoded string and returns a unsigned
        integer in at least `f` bits.
        """

        self.f = f
        self.reg = reg
        self.value = None

        if hashfunc is None:
            self.hashfunc = _hashfunc
        else:
            self.hashfunc = hashfunc

        if log is None:
            self.log = logging.getLogger("simhash")
        else:
            self.log = log

        if isinstance(value, Simhash):
            self.value = value.value
        elif isinstance(value, basestring):
            self.build_by_text(unicode(value))
        elif isinstance(value, collections.Iterable):
            self.build_by_features(value)
        elif isinstance(value, numbers.Integral):
            self.value = value
        else:
            raise Exception('Bad parameter with type {}'.format(type(value)))

    def __eq__(self, other):
        """
        Compare two simhashes by their value.

        :param Simhash other: The Simhash object to compare to
        """
        return self.value == other.value

    def _slide(self, content, width=4):
        return [content[i:i + width] for i in range(max(len(content) - width + 1, 1))]

    def _tokenize(self, content):
        content = content.lower()
        content = ''.join(re.findall(self.reg, content))
        ans = self._slide(content)
        return ans

    def build_by_text(self, content):
        features = self._tokenize(content)
        features = {k:sum(1 for _ in g) for k, g in groupby(sorted(features))}
        return self.build_by_features(features)

    def build_by_features(self, features):
        """
        `features` might be a list of unweighted tokens (a weight of 1
                   will be assumed), a list of (token, weight) tuples or
                   a token -> weight dict.
        """
        v = [0] * self.f
        masks = [1 << i for i in range(self.f)]
        if isinstance(features, dict):
            features = features.items()
        for f in features:
            if isinstance(f, basestring):
                h = self.hashfunc(f.encode('utf-8'))
                w = 1
            else:
                assert isinstance(f, collections.Iterable)
                h = self.hashfunc(f[0].encode('utf-8'))
                w = f[1]
            for i in range(self.f):
                v[i] += w if h & masks[i] else -w
        # use reversed binary str to keep the backward compatibility
        binary_str = ''.join(['0' if i <= 0 else '1' for i in v[::-1]])
        self.value = int(binary_str, 2)

    def distance(self, another):
        assert self.f == another.f
        x = (self.value ^ another.value) & ((1 << self.f) - 1)
        ans = 0
        while x:
            ans += 1
            x &= x - 1
        return ans


class SimhashIndex(object):

    def __init__(self, objs, f=64, k=2, log=None):
        """
        `objs` is a list of (obj_id, simhash)
        obj_id is a string, simhash is an instance of Simhash
        `f` is the same with the one for Simhash
        `k` is the tolerance
        """
        self.k = k
        self.f = f
        count = len(objs)

        if log is None:
            self.log = logging.getLogger("simhash")
        else:
            self.log = log

        self.log.info('Initializing %s data.', count)

        self.bucket = collections.defaultdict(set)

        for i, q in enumerate(objs):
            if i % 10000 == 0 or i == count - 1:
                self.log.info('%s/%s', i + 1, count)

            self.add(*q)

    def get_near_dups(self, simhash):
        """
        `simhash` is an instance of Simhash
        return a list of obj_id, which is in type of str
        """
        assert simhash.f == self.f

        ans = set()

        for key in self.get_keys(simhash):
            dups = self.bucket[key]
            self.log.debug('key:%s', key)
            if len(dups) > 200:
                self.log.warning('Big bucket found. key:%s, len:%s', key, len(dups))

            for dup in dups:
                sim2, obj_id = dup.split(',', 1)
                sim2 = Simhash(long(sim2, 16), self.f)

                d = simhash.distance(sim2)
                if d <= self.k:
                    ans.add(obj_id)
        return list(ans)

    def add(self, obj_id, simhash):
        """
        `obj_id` is a string
        `simhash` is an instance of Simhash
        """
        assert simhash.f == self.f

        for key in self.get_keys(simhash):
            v = '%x,%s' % (simhash.value, obj_id)
            self.bucket[key].add(v)

    def delete(self, obj_id, simhash):
        """
        `obj_id` is a string
        `simhash` is an instance of Simhash
        """
        assert simhash.f == self.f

        for key in self.get_keys(simhash):
            v = '%x,%s' % (simhash.value, obj_id)
            if v in self.bucket[key]:
                self.bucket[key].remove(v)

    @property
    def offsets(self):
        """
        You may optimize this method according to <http://www.wwwconference.org/www2007/papers/paper215.pdf>
        """
        return [self.f // (self.k + 1) * i for i in range(self.k + 1)]

    def get_keys(self, simhash):
        for i, offset in enumerate(self.offsets):
            if i == (len(self.offsets) - 1):
                m = 2 ** (self.f - offset) - 1
            else:
                m = 2 ** (self.offsets[i + 1] - offset) - 1
            c = simhash.value >> offset & m
            yield '%x:%x' % (c, i)

    def bucket_size(self):
        return len(self.bucket)



"""
FireHole
========

    FireHole is a Python package customized for data analysts.

    Website:
    Source: https://github.com/xSumner/firehole
    Bug reports: https://github.com/xSumner/firehole/issues

Using
-----

    Just write in Python

    >>> import firehole as fh
    >>> entropy = fh.Entropy()
    >>> cov = fh.COV()
    >>> ahp = fh.parse()

"""
#   Xia Tian <xsumner@hotmail.com>
#   All rights reserved.
#   Apache License.
#
# Add platform dependent shared library path to sys.path
#

from __future__ import absolute_import

import sys
if sys.version_info[:2] < (3, 5):
    m = "Python 3.5 or later is required for FireHole (%d.%d detected)"
    raise ImportError(m % sys.version_info[:2])
del sys

# These are import orderwise
from firehole.exception import *

import firehole.algorithms
from firehole.algorithms import *




# -*- coding: utf-8 -*-
"""
==========================
Evaluate and score systems
==========================
"""
import numpy as np
from firehole.exception import FireHoleNotArray
__author__ = '''Xia Tian (xsumner@hotmail.com)'''


class Entropy:
    def __init__(self):
        print("Welcome to use FireHole, this is Entropy Method !\n")
        print("For more usage, please refer to the examples folder.")

    @staticmethod
    def matrix_standardization(matrix):
        # standardize the input matrix (max-min standardization)
        standard_matrix = (matrix - matrix.min()) / (matrix.max() - matrix.min())
        return standard_matrix

    @staticmethod
    def get_entropy(matrix):
        # get the column number and row number
        m, n = matrix.shape
        k = 1 / np.log(m)
        # Add by column (axis=0: add by columns; axis=1: add by rows)
        yij = matrix.sum(axis=0)
        # pij is the prior probability of yij
        pij = matrix / yij
        # use seterr to avoid warning(for temporary)
        np.seterr(divide='ignore', invalid='ignore')
        test = pij * np.log(pij)
        np.seterr(divide='warn', invalid='warn')
        # Replace nan with 0
        test = np.nan_to_num(test)
        # Calculate the entropy of each column
        ej = -k * (test.sum(axis=0))
        # Calculate the weights
        wi = (1 - ej) / (n - np.sum(ej))

        return ej, wi

    @staticmethod
    def get_score(data_matrix, weight_matrix):
        weight_matrix = weight_matrix.copy()
        data_matrix = data_matrix.copy()
        # Calculate the final scores
        scores = weight_matrix * data_matrix
        return scores.sum(axis=0)

    def entropy(self, data):
        """Return a result dict of entropy method

        Raise an exception if the input is not numpy array

        Parameters
        ----------
        data : numpy array

        Returns
        -------
        entropy_rlt : dictionary
            A dictionary with keys of 3 strings and value of 3 numpy array

        Examples
        --------
        """
        data = data.copy()
        data_type = type(data)
        if data_type != np.ndarray:
            raise FireHoleNotArray('data', data_type)
        standard_matrix = self.matrix_standardization(data)
        entropy_value, weight = self.get_entropy(standard_matrix)
        score = self.get_score(standard_matrix, weight)
        entropy_rlt = {"entropy": entropy_value, "weight": weight, "score": score}
        return entropy_rlt


class COV:
    def __init__(self):
        print("Welcome to use FireHole, this is coefficient of variation method !\n")
        print("For more usage, please refer to the examples folder. ")

    @staticmethod
    def coefficient_of_variation(data):
        data_type = type(data)
        if data_type != np.ndarray:
            raise FireHoleNotArray('data', data_type)
        # standard deviation divided by mean (without nan value)
        v = np.nanstd(data, axis=0) / np.nanmean(data, axis=0)
        return v

    def comprehensive(self, data, target, pos_neg=1, decimals=4):
        """Return a result dict of coefficient  method

        Raise an exception if the input is not numpy array

        Parameters
        ----------
        data : numpy array
        target : numpy array
        pos_neg : numpy array
        decimals: int

        Returns
        -------
        cov_rlt : numpy array
            A numpy array of the final score

        Examples
        ----
        """
        # check the target data type
        target_type = type(target)
        if target_type != np.ndarray:
            raise FireHoleNotArray('target', target_type)
        # check the input size to avoid Broadcasting Error !
        data_shape = set(data.shape)
        target_shape = set(target.shape)
        shape_set = data_shape & target_shape
        if shape_set == set():
            raise Exception("data and target should have at least one same size !")
        if [pos_neg != 1]:
            pos_neg_type = type(pos_neg)
            if pos_neg_type != np.ndarray:
                raise FireHoleNotArray('pos_neg', pos_neg_type)
            pos_neg_shape = set(pos_neg.shape)
            if pos_neg_shape & shape_set == set():
                raise Exception("pos_neg has wrong size !")
        # calculate the coefficient of variation
        v = self.coefficient_of_variation(data)
        # calculate the weight
        w = v / v.sum()
        # use the filter array to filter the index
        pos = (pos_neg + 1) / 2
        neg = (1 - pos_neg) / 2
        # calculate the compliance of positive index
        g_pos = np.divide(data, target)
        g_pos = g_pos * pos
        # calculate the compliance of negative index
        g_neg = np.divide(target, data)
        g_neg = g_neg * neg
        # add to get the final compliance matrix
        g = g_pos + g_neg
        # convert nan into zero to avoid error
        g = np.nan_to_num(g)
        # calculate the final result
        cov_rlt = (w * g).sum(axis=1)
        # round the result according to the decimals
        rlt = np.around(cov_rlt, decimals)
        return rlt



# -*- coding: utf-8 -*-
"""ahp.parser

This module contains the functions required to perform AHP Model validation.
"""

from queue import Queue
import numpy as np

from firehole.algorithms.ahp.errors import *
from firehole.algorithms.ahp.hierarchy import AHPModel
from firehole.algorithms.ahp.methods import *

methods = ('approximate', 'eigenvalue', 'geometric')


def _type(val):
    return val.__class__.__name__


def _check_ahp_list(name, value):
    if not isinstance(value, list):
        raise AHPTypeMismatchError(name, 'list', _type(value))

    if not value:
        raise AHPFieldEmptyError(name)

    for elem in value:
        if not isinstance(elem, str):
            raise AHPTypeMismatchError(name, 'str', _type(elem), list_elements=True)

    if len(value) != len(set(value)):
        raise AHPContainsDuplicateError(name)


def _check_ahp_preference_matrix(name, p_m, kind, length):
    if p_m is None:
        raise AHPMissingPreferenceMatrixError(kind, name)

    p_m = np.array(p_m)

    width, height = p_m.shape
    if width != height or width != length:
        raise AHPNonSquarePreferenceMatrixError(kind, name, length, width, height)


def validate_model(model):
    """Validate the passed AHP model.

    Args:
        model (dict): The Analytic Hierarchy Process model.

    Raises:
        AHPModelError when the model validation fails.
    """

    if not isinstance(model, dict):
        raise AHPTypeMismatchError('model', 'dict', _type(model))

    if not model:
        raise AHPModelError('AHP Model cannot be an empty dictionary or json object')

    try:
        method = model['method']
    except KeyError:
        raise AHPFieldEmptyError('method')

    if not isinstance(method, str):
        raise AHPTypeMismatchError('method', 'str', _type(method))

    if method not in methods:
        raise AHPMethodUnsupportedError((', '.join(methods)), method)

    _check_ahp_list('criteria', model['criteria'])
    _check_ahp_list('alternatives', model['alternatives'])

    n_alternatives = len(model['alternatives'])
    preference_matrices = model['preferenceMatrices']
    criteria = model['criteria']
    criteria_queue = Queue()

    criteria_p_m = preference_matrices.get('criteria')
    _check_ahp_preference_matrix(name='criteria', p_m=criteria_p_m, kind="criteria", length=len(criteria))

    for criterion in criteria:
        criteria_queue.put(criterion)

    sub_criteria_map = model.get('subCriteria')

    while not criteria_queue.empty():
        criterion = criteria_queue.get()
        sub_criteria = sub_criteria_map.get(criterion)

        if sub_criteria:
            _check_ahp_list('subCriteria:{}'.format(criterion), sub_criteria)

            p_m = preference_matrices.get('subCriteria:{}'.format(criterion))
            _check_ahp_preference_matrix(name=criterion, p_m=p_m, kind="subCriteria", length=len(sub_criteria))

            for sub_criterion in sub_criteria:
                criteria_queue.put(sub_criterion)
        else:
            p_m = preference_matrices.get('alternatives:{}'.format(criterion))
            _check_ahp_preference_matrix(name=criterion, p_m=p_m, kind="alternatives", length=n_alternatives)


def parse(model):
    """Parse the passed AHP model.

    Args:
        model (dict): The Analytic Hierarchy Process model.

    Returns:
        AHPModel with the specified solver.
    """
    validate_model(model)

    method_name = model['method'].capitalize() + 'Method'
    solver = globals()[method_name]

    return AHPModel(model, solver)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
**********
Exceptions
**********

Base exception and errors for FireHole

"""
__author__ = '''Xia Tian (xsumner@hotmail.com)'''


# the root of all Exceptions
class FireHoleException(Exception):
    """Base class for exceptions in FireHole"""


class FireHoleAlgorithmError(FireHoleException):
    """Exception for unexpected termination of algorithms."""


class FireHoleUnfeasible(FireHoleAlgorithmError):
    """Exception raised by algorithm trying to solve a problem
    instance that has no feasible solution"""


class FireHoleNotArray(FireHoleAlgorithmError):
    """Exception raised by introducing data which is not a numpy
    n-dimension array"""
    def __init__(self, arg_name, data_type):
        self.name = arg_name
        self.type = str(data_type)

    def __str__(self):
        print("{} is {}, not a numpy array.".format(self.name, self.type))



# -*- coding: utf-8 -*-
"""ahp.methods.geometric

This module contains the class implementing the geometric priority estimation method.
"""

import numpy as np

from firehole.algorithms.ahp.methods import Method


class GeometricMethod(Method):
    """Geometric priority estimation method
    """

    def estimate(self, preference_matrix):
        super()._check_matrix(preference_matrix)
        height = preference_matrix.shape[1]
        vec = np.prod(preference_matrix, axis=1) ** 1/height
        return vec / np.sum(vec)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Setup script for FireHole

You can install FireHole with

python setup.py install
"""
from setuptools import setup
from os import path

DIR = path.dirname(path.abspath(__file__))

DESCRIPTION = "FireHole (fh) is a Python package for Data Analyst."

AUTHORS = 'Xia Tian'

URL = 'https://github.com/xSumner/firehole'

EMAIL = 'xsumner@hotmail.com'

with open(path.join(DIR, 'requirements.txt')) as f:
    INSTALL_PACKAGES = f.read().splitlines()

with open(path.join(DIR, 'README.md'), encoding='UTF-8') as f:
    README = f.read()


VERSION = "0.4.0"

packages = ["firehole",
            "firehole.algorithms",
            "firehole.algorithms.ahp",
            "firehole.algorithms.ahp.hierarchy",
            "firehole.algorithms.ahp.methods",
            "firehole.algorithms.flashtext",
            "firehole.algorithms.similarity"]

# add the tests
package_data = {
    "firehole": ['tests/*.py'],
    "firehole.firehole": ['tests/*.py']
    }


setup(
    name='firehole',
    packages=packages,
    description=DESCRIPTION,
    long_description=README,
    long_description_content_type='text/markdown',
    install_requires=INSTALL_PACKAGES,
    version=VERSION,
    url=URL,
    author=AUTHORS,
    author_email=EMAIL,
    keywords=['Data Analysis', 'AHP'],
    tests_require=['pytest'],
    package_data=package_data,
    include_package_data=True,
    python_requires='>=3'
)



# -*- coding: utf-8 -*-
"""ahp.hierarchy.ahpcriterion

This module contains the class definition for the AHP Criteria Node in the hierarchy model.
"""

import numpy as np

from firehole.algorithms.ahp.methods import EigenvalueMethod
from firehole.algorithms.ahp.utils import normalize_priorities


class AHPCriterion:
    """AHPCriterion

    Args:
        name (str): Name of the criterion this node resembles.
        model (dict): The Analytic Hierarchy Process model.
        solver (ahp.methods): Method used when calculating the priorities of the lower layer.
    """

    def __init__(self, name, model, solver=EigenvalueMethod):
        self.name = name
        self.solver = solver()
        self.preference_matrices = model['preferenceMatrices']
        self.leaf = False

        sub_criteria = model['subCriteria'].get(name)

        if sub_criteria is not None:
            self.sub_criteria = [AHPCriterion(n, model, solver) for n in sub_criteria]
            self.p_m_key = 'subCriteria:{}'.format(name)
        else:
            self.leaf = True
            self.p_m_key = 'alternatives:{}'.format(name)

    def __str__(self):
        return self.name

    def get_priorities(self):
        """Get the priority of the nodes in the level below this node.

        Returns:
            Priorities at current level, normalized if an internal node.
        """
        p_m = np.array(self.preference_matrices[self.p_m_key])
        sub_crit_pr = self.solver.estimate(p_m)

        if self.leaf:
            return sub_crit_pr

        criteria_pr = [criterion.get_priorities() for criterion in self.sub_criteria]

        return normalize_priorities(criteria_pr, sub_crit_pr)



# -*- coding: utf-8 -*-
"""ahp.methods

This module contains the imports for functions, classes and constants exported.
"""

from .method import Method
from .approximate import ApproximateMethod
from .eigenvalue import EigenvalueMethod
from .geometric import GeometricMethod



# -*- coding: utf-8 -*-
"""ahp.methods.approximate

This module contains the class implementing the geometric priority estimation method.
"""

import numpy as np

from firehole.algorithms.ahp.methods import Method


class ApproximateMethod(Method):
    """Approximate priority estimation method
    """

    def estimate(self, preference_matrix):
        super()._check_matrix(preference_matrix)

        row_sums = np.sum(preference_matrix, axis=1)
        total_sum = np.sum(row_sums)

        return row_sums / total_sum



# -*- coding: utf-8 -*-
"""ahp.utils

This module contains the common functions and methods used by other modules in the class.
"""

import numpy as np


def normalize_priorities(criteria_pr, global_pr):
    """Normalize the priorities received from the lower layer.

    This function performs a Global Prioritization at the current node.

    Args:
        criteria_pr (list(list(float))): The priorities of all the alternatives/sub-criteria.
        global_pr (list(float)): The global priorities of all criteria.

    Returns:
       Normalized priorities
    """
    
    return np.dot(global_pr, criteria_pr)
